!Module storing observed datasets, their points and window functions
!and routines for computing the likelihood
!Cls passed to these routines are in in MicroK^2, no 2pi or l(l+1)
!WMAP data is handled separately as a special case

!If windows_are_bandpowers=T Windows functions read in from files are assumed to be W_l/l. 
!(unless windows_are_bare=T, in which case we assume they are W_l)
!The band powers are obtained from
!W_l by \sum_l (l+1/2) W_l C_l/2pi where W_l is normalized so \sum_l (l+1/2)/l(l+1) W_l = 1.
!If windows_are_normalized=T  the windows are assumed to be already normalized; 
!this must be the case for polarization in which the window files contain
! l TT TE EE BB
!contributions to each band power. Usually all but two columns will be zeros.

!If windows_are_bandpowers=F then window functions are assumed raw, and directly related
!to the bandpowers by B_i=sum_l W_{il} C_l

!This code is very indebted to Sarah's cosmarg
!Analytic marginalization over calibration and beam from astro-ph/0112114

!x factors are described in astro-ph/9808264. Essentially they are a first correction
!for the bandpower errors being non-Gaussian. If no x-factors we assume they are gaussian.
!Using x-factors we transform to a variable Z = log (bandpower + x), and assume Z has Gaussian errors.
!When x-factors are used the inverse covariance matrix is assumed to be that for Z.
!See the RADPACK page for data http://bubba.ucdavis.edu/~knox/radpack.html

!This version March 2004
!Mar 04: added first_band parameter to .dataset files, added format for doing exact likelihoods


module cmbdata
use settings
use cmbtypes
implicit none

 
!if CMBdataset%has_xfactors then obs, var and N_inv are for the offset-lognormal variables Z

  Type CMBdatapoint
    integer win_min, win_max
     !Ranges in which window is non-zero
    real, pointer, dimension(:,:) :: window
     !Normalized window function in l
    real obs, err_minus, err_plus, sigma, var
     !Observed value of l(l+1) C_l/2pi bandpowers in MicroK^2, with errors
    real beam_err !fractional beam error (file is value in MicroK^2)
    logical inc_pol

  end Type CMBdatapoint

  Type CMBdataset
    logical :: use_set
    logical :: has_pol, windows_are_bandpowers,windows_are_normalized
    real :: calib_uncertainty
    logical :: beam_uncertain, has_corr_errors, has_xfactors
    integer :: num_points, file_points
    character(LEN=20) :: name
    Type(CMBdatapoint), pointer, dimension(:) :: points
    real, pointer, dimension(:,:) :: N_inv
    real, pointer, dimension(:) :: xfactors
    logical :: all_l_exact
    integer :: all_l_lmax
    real, pointer, dimension(:,:) :: all_l_obs, all_l_noise
    real, pointer, dimension(:) :: all_l_fsky
  end Type CMBdataset

  integer :: num_datasets = 0
  Type(CMBdataset) datasets(10)

  logical :: init_MAP = .true.
 
  integer, parameter :: halfsteps = 5 !do 2*halfsteps+1 steps in numerical marginalization
  real margeweights(-halfsteps:halfsteps)
  real :: margenorm = 0 
  private halfsteps, margeweights, margenorm
contains


 subroutine ReadWindow(AP, aname, are_bare, aset)
   Type(CMBdatapoint) AP
   character(LEN=*), intent(IN) :: aname
   logical, intent(IN) :: are_bare
   Type (CMBdataset) :: aset
   integer l, ncls
   real wpol(1:num_cls-1),ll, IW,w
   character(LEN=200) tmp

   if (Feedback > 1) write (*,*) 'reading window: '//trim(aname)

   if (aset%has_pol) then
    ncls = num_cls
   else
    ncls = 1
   endif
   allocate(AP%window(ncls,lmax))

   AP%window = 0

   call OpenTxtFile(aname, 50)

     do
     if (aset%has_pol) then
       read(50,'(a)',end=1) tmp
       read(tmp,*, end=1) ll, w, wpol
     else
       read(50,*, end=1) ll, w
     end if

     l=nint(ll)
     if (abs(l-ll) > 1e-4) stop 'non-integer l in window file' 
     if (l>=2 .and. l<=lmax) then
        AP%window(1,l) = w
        if(aset%has_pol) then
         AP%window(2:num_cls,l) = wpol
        end if
       if (.not. are_bare) AP%window(:,l) = AP%window(:,l)*l
     else
        if (l>lmax .and. w /= 0) then
           write (*,*) 'Warning: Window function non-zero outside l_max: ',trim(aname)
           write (*,*) 'assuming contribution is negligible'
           exit
        end if
     end if

     end do
1  close(50)

   do l=2, lmax
      if (any(AP%window(:,l)/=0)) then
         AP%win_min = l
         exit
      end if
   end do

   do l=lmax,2,-1
      if (any(AP%window(:,l)/=0)) then
         AP%win_max = l
         exit
      end if
   end do

   AP%inc_pol = .false.

   if (aset%windows_are_bandpowers) then
       IW = 0
       do l = AP%win_min, AP%win_max
          IW = IW + ((l+0.5)*AP%window(1,l))/(l*(l+1))
          AP%window(:,l) = AP%window(:,l)*(l+0.5)/(2*pi)
       end do
       if (.not. aset%windows_are_normalized) then
          if (aset%has_pol) stop 'Polarization windows must be normalized'
          AP%window(1,AP%win_min:AP%win_max) = &
                      AP%window(1,AP%win_min:AP%win_max)/IW
       end if
   end if
    !If has_pol we are assuming windows are normalized
  
   if (aset%has_pol) AP%inc_pol = any(AP%window(2:num_cls,:)/=0)
  
 end subroutine ReadWindow

 subroutine ReadAllExact(aset)
      Type (CMBdataset) :: aset
      character(LEN=120) :: fname
      integer l, idum
    !In this case we have data for every l, and use exact full-sky likelihood expression
    !with some fudge factor fsky_eff^2 to reduce the degrees of freedom: fsky^eff*(2l+1)

       aset%num_points = 0
       aset%all_l_lmax = Ini_Read_Int('all_l_lmax')
       if (aset%all_l_lmax > lmax) stop 'cmbdata.f90: all_l_lmax > lmax'
       if (aset%has_pol) then
        allocate(aset%all_l_obs(2:aset%all_l_lmax,num_cls))
        allocate(aset%all_l_noise(2:aset%all_l_lmax,2))
       else
        allocate(aset%all_l_obs(2:aset%all_l_lmax,1))
        allocate(aset%all_l_noise(2:aset%all_l_lmax,1))
       end if
       allocate(aset%all_l_fsky(2:aset%all_l_lmax))
       fname = Ini_Read_String('all_l_file')
       call OpenTxtFile(fname,tmp_file_unit)
!File format:
!l C_TT (C_TE C_EE [C_BB]) N_T (N_P) fsky_eff
!Must have num_cls set to correct value for file
       do l = 2, aset%all_l_lmax
          read (tmp_file_unit, *, end=100, err=100) idum, aset%all_l_obs(l,:), aset%all_l_noise(l,:), aset%all_l_fsky(l)
          if (idum /= l) stop 'Error reading all_l_file'
        end do
       close(tmp_file_unit)

       return
100    stop 'Error reading all_l_file file'
     

 end subroutine ReadAllExact


 function ChiSqExact(cl, aset) 
  !Compute -ln(Likelihood) 
   real cl(lmax,num_cls)
   Type(CMBdataset) :: aset
   integer l
   real ChiSqExact, chisq, term, CT, CE, CB
   real dof

   chisq=0

   do l=2, aset%all_l_lmax
     dof = aset%all_l_fsky(l)**2*(2*l+1)
       !Ignoring l correlations but using f_sky^2_eff fudge factor may be a good approx
       !for nearly full sky observations
     CT = cl(l,1) + aset%all_l_noise(l,1)
     if (aset%has_pol) then
      CE = cl(l,3) + aset%all_l_noise(l,2)
      term = CT*CE- cl(l,2)**2
      chisq = chisq + dof*( &
       (CT*aset%all_l_obs(l,3) + CE*aset%all_l_obs(l,1) - 2 *cl(l,2)*aset%all_l_obs(l,2))/term &
        + log( term/ (aset%all_l_obs(l,1)*aset%all_l_obs(l,3) - aset%all_l_obs(l,2)**2)) -2)
      if (num_cls>3) then
        !add in BB
        CB = cl(l,num_cls) + aset%all_l_noise(l,2)
        chisq = chisq + dof * (aset%all_l_obs(l,num_cls)/CB &
                +log(CB/aset%all_l_obs(l,num_cls)) - 1)
      end if
     else
        chisq = chisq + dof * (aset%all_l_obs(l,1)/CT &
                +log(CT/aset%all_l_obs(l,1)) - 1)
     end if
   end do
   ChiSqExact = ChiSq

 end function ChiSqExact

 subroutine ReadDataset(aname)
   character(LEN=*), intent(IN) :: aname
   character(LEN=120) :: InLine, window_dir, Ninv_file, xfact_file, band_file
   logical bad, windows_are_bare
   Type (CMBdataset) :: aset
   integer i, first_band, use_i
   real, pointer, dimension(:,:) :: tmp_mat
   real, pointer, dimension(:) :: tmp_arr

   num_datasets = num_datasets + 1
   if (num_datasets > 10) stop 'too many datasets'

   if (aname == 'MAP' .or. aname == 'WMAP') then 
     aset%name = 'WMAP'
     datasets(num_datasets) = aset
     return   
   end if

   call Ini_Open(aname, 1, bad, .false.)
   if (bad) then
     write (*,*)  'Error opening dataset file '//trim(aname)
     stop
   end if
   Ini_fail_on_not_found = .true.
      
   aset%name = Ini_Read_String('name')
   aset%use_set =.true.
    
   if (Feedback > 0) write (*,*) 'reading: '//trim(aset%name)

   Ini_fail_on_not_found = .false.

   aset%has_pol = Ini_Read_Logical('has_pol',.false.)
   
   aset%all_l_exact = Ini_Read_Logical('all_l_exact',.false.)
   if (aset%all_l_exact) then
       call ReadAllExact(aset)
   else
    !Otherwise do usual guassian/offset lognormal stuff

       aset%num_points = Ini_Read_Int('num_points')

       aset%calib_uncertainty = Ini_Read_Double('calib_uncertainty')
       aset%beam_uncertain = Ini_Read_logical('beam_uncertainty')

        
       window_dir  = Ini_Read_String('window_dir')

       windows_are_bare = Ini_Read_Logical('windows_are_bare',.false.)
       aset%windows_are_bandpowers = Ini_Read_Logical('windows_are_bandpowers',.true.)
       aset%windows_are_normalized = Ini_Read_Logical('windows_are_normalized',.false.)
  
       aset%file_points = Ini_read_Int('file_points',aset%num_points)
       first_band = Ini_read_Int('first_band',1)
       if (first_band + aset%num_points > aset%file_points+1) then
            write (*,*)  'Error with dataset file '//trim(aname)
            write (*,*) 'first_band + num_points > file_points'
            stop
       end if
    !Read in the observed values, errors and beam uncertainties
       allocate(aset%points(aset%num_points))
       band_file = Ini_Read_String('bandpowers')
       if (band_file /= '') call OpenTxtFile(band_file, 51)
       Ini_fail_on_not_found = .true.
       do i=1, aset%num_points + first_band -1
          if (band_file /= '') then
            read(51,'(a)') InLine
          else
            InLine = Ini_Read_String(numcat('data',i))
          end if
          if (i < first_band) cycle
          use_i = i - first_band + 1          
          if (aset%beam_uncertain) then
            read(InLine,*) aset%points(use_i)%obs, aset%points(use_i)%err_minus, aset%points(use_i)%err_plus, &
                            aset%points(use_i)%beam_err
            aset%points(use_i)%beam_err = aset%points(use_i)%beam_err/aset%points(use_i)%obs
          else
           read(InLine,*) aset%points(use_i)%obs, aset%points(use_i)%err_minus, aset%points(use_i)%err_plus
           aset%points(use_i)%beam_err = 0
          end if
          aset%points(use_i)%sigma = (aset%points(use_i)%err_minus + aset%points(use_i)%err_plus)/2
          aset%points(use_i)%var = aset%points(use_i)%sigma**2
          call ReadWindow(aset%points(use_i),trim(window_dir)//'/'//trim(numcat(aset%name,i)),windows_are_bare,aset)
       end do
       if (band_file /= '') Close(51)
  

    !See if the inverse covariance matrix is given (otherwise assume diagonal)
       Ini_fail_on_not_found = .false.
    
       Ninv_file = Ini_Read_String('N_inv')
       aset%has_corr_errors = Ninv_file /= ''
       if (aset%has_corr_errors) then
          allocate(tmp_mat(aset%file_points,aset%file_points))
          allocate(aset%N_inv(aset%num_points,aset%num_points))
          call ReadMatrix(Ninv_file, tmp_mat,aset%file_points,aset%file_points)
          if (aset%num_points /= aset%file_points) then
            !!Changed to truncation of covariance matrix, AL: May 2003
            call Matrix_Inverse(tmp_mat)
            aset%N_inv = tmp_mat(first_band:first_band+aset%num_points-1,&
                    first_band:first_band + aset%num_points -1)
            call Matrix_Inverse(aset%N_inv)
          else
            aset%N_inv = tmp_mat(1:aset%num_points,1:aset%num_points)
          end if
          deallocate(tmp_mat)
       end if

       if (Ini_Read_Logical('use_hyperparameter',.false.)) stop 'Hyperparameters deprecated'

    !See if xfactors are given

       xfact_file = Ini_Read_String('xfactors')
       aset%has_xfactors = xfact_file /= ''

       if (aset%has_xfactors) then
          allocate(tmp_arr(aset%num_points + first_band -1))
          call ReadVector(xfact_file, tmp_arr,aset%num_points+ first_band -1)
          allocate(aset%xfactors(aset%num_points))
          aset%xfactors = tmp_arr(first_band:first_band+aset%num_points-1)
          deallocate(tmp_arr)
          aset%points(:)%var = aset%points(:)%var/(aset%points(:)%obs +aset%xfactors)**2
          aset%points(:)%obs = log(aset%points(:)%obs +aset%xfactors)
        
       end if

   end if !not all_l_exact
   
   call Ini_Close

   datasets(num_datasets) = aset

 end subroutine ReadDataset

 function GetWinBandPower(AP, cl)
   real  GetWinBandPower
   real cl(lmax,num_cls)
   Type(CMBdatapoint) AP
   integer l
   real bandpower

   bandpower = sum(cl(AP%win_min:AP%win_max,1)*AP%window(1,AP%win_min:AP%win_max))

   if (AP%inc_pol) then
       do l= AP%win_min, AP%win_max
           bandpower = bandpower + sum(cl(l,2:num_cls)*AP%window(2:num_cls,l))
       end do
   endif
   GetWinBandPower  = bandpower
     
 end function GetWinBandPower
 
 subroutine InitNumericalMarge
   integer i

   do i= -halfsteps, halfsteps
    margeweights(i) = exp(-(i*3/real(halfsteps))**2/2)
   end do
   margenorm = sum(margeweights)

 end subroutine InitNumericalMarge

 function GetCalibMargexChisq(bandpowers, aset)
  !Numerically integrate over the calibration uncertainty
  !Assume Gaussian prior, as for analytic calculation without x-factors
  !Could also Monte-Carlo
   real GetCalibMargexChisq
   Type(CMBdataset) aset
   real bandpowers(aset%num_points),beambandpowers(aset%num_points),diffs(aset%num_points)
   real calib, chisq(-halfsteps:halfsteps),chisqcalib(-halfsteps:halfsteps)
   real minchisq
   integer i, ibeam

   if (margenorm == 0) call InitNumericalMarge

   do ibeam= -halfsteps, halfsteps

      if (aset%beam_uncertain) then
         beambandpowers = bandpowers*(1 + aset%points(:)%beam_err*ibeam*3/real(halfsteps))!Go out to 3 sigma
      else
         beambandpowers = bandpowers
      end if

      do i=-halfsteps,halfsteps

         calib = 1 + aset%calib_uncertainty*i*3./halfsteps !Go out to 3 sigma
         diffs(:) = aset%points(:)%obs- log(calib*beambandpowers(:) + aset%xfactors(:))
         if (aset%has_corr_errors) then
            chisq(i) = SUM(diffs*MATMUL(aset%N_inv,diffs))
          else
            chisq(i) = SUM(diffs**2/aset%points(:)%var)  
         end if
      end do

      minchisq = minval(chisq)

      chisqcalib(ibeam) = -2*log(sum(margeweights*exp(max(-30.,-(chisq-minchisq)/2)))/margenorm) + minchisq

      if (.not. aset%beam_uncertain) then
         GetCalibMargexChisq = chisqcalib(ibeam)
         return
      end if

   end do

   minchisq = minval(chisqcalib)
   GetCalibMargexChisq = -2*log(sum(margeweights*exp(max(-30.,-(chisqcalib-minchisq)/2)))/margenorm) + minchisq
  
 end function GetCalibMargexChisq


 function CalcLnLike(cl, aset) 
  !Compute -ln(Likelihood) 
   real cl(lmax,num_cls), CalcLnLike
   Type(CMBdataset) aset
   integer i
   real chisq
   real chi2op, chi2pp, wpp, wdd
   real chi2dd,chi2pd,chi2od
   real bandpowers(aset%num_points), diffs(aset%num_points), tmp(aset%num_points), beam(aset%num_points)
   real denom

   if (.not. aset%use_set) then
      CalcLnLike = 0
      return
   end if

   if (aset%all_l_exact) then
     
     chisq = ChiSqExact(cl,aset)
    
   else
   
       denom = 1 !Assume Prob \propto exp(-chisq/2)/sqrt(denom)

       do i=1, aset%num_points
          bandpowers(i) = GetWinBandPower(aset%points(i), cl)
       end do
   
       if (aset%has_xfactors .and. (aset%calib_uncertainty > 1e-4 .or. aset%beam_uncertain))  then

          chisq = GetCalibMargexChisq(bandpowers,aset)

       else

        do i=1, aset%num_points
          if (aset%has_xfactors) then
           !obs in this case is Z = log(observed + x)
           diffs(i) = aset%points(i)%obs- log(bandpowers(i) + aset%xfactors(i))
          else
           diffs(i) = aset%points(i)%obs - bandpowers(i)
          end if
       end do

       if (aset%has_corr_errors) then
          chisq = SUM(diffs*MATMUL(aset%N_inv,diffs))
       else
          chisq = SUM(diffs**2/aset%points(:)%var)  
       end if

       if (aset%calib_uncertainty > 1e-4 .or. aset%beam_uncertain) then
 
         if (aset%has_corr_errors) then
            tmp = MATMUL(aset%N_inv,bandpowers)
         else
            tmp = bandpowers/aset%points(:)%var  
         end if
         chi2op = SUM(diffs*tmp)
         chi2pp = SUM(bandpowers*tmp)
         if (aset%beam_uncertain) then
             beam = aset%points(:)%beam_err*bandpowers
             if (aset%has_corr_errors) then
                tmp = MATMUL(aset%N_inv,beam)
             else
                tmp = beam/aset%points(:)%var 
             end if
             chi2dd = SUM(beam*tmp)
             chi2pd = SUM(bandpowers*tmp)
             chi2od = SUM(diffs*tmp)
         end if

         if (aset%calib_uncertainty > 1e-4) then
          !analytic marginalization over calibration uncertainty
          wpp = 1/(chi2pp+1/aset%calib_uncertainty**2) 
          chisq = chisq - wpp*chi2op**2 
          denom = denom/wpp*aset%calib_uncertainty**2
         else
          wpp = 0
         end if

         if (aset%beam_uncertain) then
          !analytic marginalization over beam uncertainty  
          wdd=1/(chi2dd-wpp*chi2pd**2+1)
          chisq = chisq - wdd*(chi2od-wpp*chi2op*chi2pd)**2
          denom = denom/wdd
         end if

       end if

       end if

        if (denom /= 1) chisq = chisq + log(denom)
  
    end if
    CalcLnLike = chisq/2
  
 end function CalcLnLike

 function CMBLnLike(cl)
  real cl(lmax,num_cls), CMBLnLike
  integer i
  real tot(num_datasets)
 
  do i=1, num_datasets
     if (datasets(i)%name == 'WMAP') then
      tot(i) = MAPLnLike(cl)
     else
      tot(i) = CalcLnLike(cl,datasets(i))
     end if
  end do
  CMBLnLike = SUM(tot) 
 end function


 function MAPLnLike(cl)
  use WMAP
  real cl(lmax,num_cls), MAPLnLike
  integer  l
  real(WMAP_precision) clTT(WMAP_lmax_TT), clTE(WMAP_lmax_TT), ClEE(WMAP_lmax_TT)
  integer stat 
  character(LEN=20) :: ttFile = 'WMAP/tt_diag.dat'
  character(LEN=20) :: ttOffDiag ='WMAP/tt_offdiag.dat'
  character(LEN=20) :: teFile   = 'WMAP/te_diag.dat'
  character(LEN=20) :: teOffDiag ='WMAP/te_offdiag.dat'
 
  if (Init_MAP) then
   if (lmax<WMAP_lmax_TT) stop 'lmax not large enough for WMAP'
   if (Feedback>0) write(*,*) 'reading WMAP data'
   Call WMAP_init(ttFile, ttOffDiag, teFile, teOffDiag, stat)

   if (stat /=0) stop 'Error reading WMAP files'
   if (Feedback>0) write(*,*) 'WMAP read'
   Init_MAP = .false.
  end if

  do l = 2, WMAP_lmax_TT
      clTT(l) = cl(l,1)*l*(l+1)/twopi
      clTE(l) = cl(l,2)*l*(l+1)/twopi
      clEE(l) = cl(l,3)*l*(l+1)/twopi
  end do

  MAPLnLike = -(WMAP_LnLike_TT(clTT) + WMAP_LnLike_TE(clTT, clTE, clEE))

 end function


end module cmbdata


